@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
-# Copyright (c) 2022-2023 Salvador E. Tropea
-# Copyright (c) 2022-2023 Instituto Nacional de Tecnologïa Industrial
+# Copyright (c) 2022-2024 Salvador E. Tropea
+# Copyright (c) 2022-2024 Instituto Nacional de Tecnologïa Industrial
 # License: Apache 2.0
 # Project: KiAuto (formerly kicad-automation-scripts)
 import atexit
@@ -10,34 +10,42 @@ import psutil
 from queue import Queue, Empty
 import re
 import shutil
-from sys import exit
+from sys import exit, exc_info
 from tempfile import mkdtemp
 from threading import Thread
 import time
+from traceback import extract_stack, format_list, print_tb
 from kiauto.misc import KICAD_DIED, CORRUPTED_PCB, PCBNEW_ERROR, EESCHEMA_ERROR
 from kiauto import log
 from kiauto.ui_automation import xdotool, wait_for_window, wait_point, text_replace
+from kiauto.file_util import wait_for_file_created_by_process
 
 KICAD_EXIT_MSG = '>>exit<<'
 INTERPOSER_OPS = 'interposer_options.txt'
 IGNORED_DIALOG_MSGS = {'The quick brown fox jumps over the lazy dog.', '0123456789'}
 BOGUS_FILENAME = '#'
+KIKIT_HIDE = 'Specify which components to hide'
 # These dialogs are asynchronous, they can pop-up at anytime.
 # One example is when the .kicad_wks is missing, KiCad starts drawing and then detects it.
 INFO_DIALOGS = {'KiCad PCB Editor Information', 'KiCad Schematic Editor Information'}
+WARN_DIALOGS = {'KiCad PCB Editor Warning', 'KiCad Schematic Editor Warning'}
+ASYNC_DIALOGS = INFO_DIALOGS | WARN_DIALOGS
 
 
 def check_interposer(args, logger, cfg):
     # Name of the interposer library
-    interposer_lib = os.path.abspath(os.path.join(os.path.dirname(__file__), 'interposer', 'libinterposer.so'))
+    machine = platform.machine().lower()
+    extra_name = '' if machine == 'x86_64' else '_'+machine
+    interposer_lib = os.path.abspath(os.path.join(os.path.dirname(__file__), 'interposer', f'libinterposer{extra_name}.so'))
+    logger.debug(f'Looking for interposer lib: {interposer_lib}')
     if (not os.path.isfile(interposer_lib) or  # The lib isn't there
        args.disable_interposer or              # The user disabled it
        os.environ.get('KIAUTO_INTERPOSER_DISABLE') or  # The user disabled it using the environment
-       platform.system() != 'Linux' or 'x86_64' not in platform.platform()):  # Not Linux 64 bits x86
+       platform.system() != 'Linux'):  # Not Linux
         interposer_lib = None
     else:
         os.environ['LD_PRELOAD'] = interposer_lib
-        logger.debug('** Using interposer: '+interposer_lib)
+        logger.debug('** Interposer lib found')
     cfg.use_interposer = interposer_lib
     cfg.enable_interposer = interposer_lib or args.interposer_sniff
     cfg.logger = logger
@@ -57,8 +65,9 @@ def dump_interposer_dialog(cfg):
                 cfg.interposer_dialog.append('>>Interposer<<:{} (@{} D {})'.format(line[:-1], round(tm, 3), round(diff, 3)))
         except Empty:
             pass
-    for ln in cfg.interposer_dialog:
-        cfg.flog_int.write(ln+'\n')
+    if hasattr(cfg, 'interposer_dialog'):
+        for ln in cfg.interposer_dialog:
+            cfg.flog_int.write(ln+'\n')
     cfg.flog_int.close()
 
 
@@ -186,16 +195,22 @@ def wait_queue(cfg, strs='', starts=False, times=1, timeout=300, do_to=True, kic
             cfg.logger.debug('Interposer match, times='+str(times))
         if (not with_windows and not kicad_can_exit and line.startswith('GTK:Window Title:') and
             # The change in the unsaved status is ignored here
-           not(not cfg.ki5 and line.endswith(cfg.window_title_end))):
+           not (not cfg.ki5 and line.endswith(cfg.window_title_end))):
             # We aren't expecting a window, but something seems to be there
             # Note that window title change is normal when we expect KiCad exiting
             title = line[17:]
             if title in INFO_DIALOGS:
                 # Async dialogs
                 dismiss_pcb_info(cfg, title)
-            elif title == 'pcbnew Warning':
+            elif title == 'pcbnew Warning' or title in WARN_DIALOGS:
                 # KiCad 5 error during post-load, before releasing the CPU
+                # KiCad 7 missing fonts
                 dismiss_pcbnew_warning(cfg, title)
+            elif cfg.ki8 and title == 'KiCad PCB Editor Error':
+                dismiss_pcbnew_error(cfg, title)
+            elif title.startswith(KIKIT_HIDE):
+                # Buggy KiKit plugin creating a dialog at start-up (many times)
+                pass
             else:
                 unknown_dialog(cfg, title)
             if dialog_interrupts:
@@ -245,11 +260,14 @@ def wait_kicad_ready_i(cfg, swaps=0, kicad_can_exit=False):
     return res
 
 
-def open_dialog_i(cfg, name, keys, no_show=False, no_wait=False, no_main=False, extra_msg=None):
+def open_dialog_i(cfg, name, keys, no_show=False, no_wait=False, no_main=False, extra_msg=None, raise_if=None):
     wait_point(cfg)
     # Wait for KiCad to be sleeping
     wait_kicad_ready_i(cfg)
     cfg.logger.info('Opening dialog `{}` {}'.format(name, '('+extra_msg+')' if extra_msg is not None else ''))
+    # cfg.logger.error('Durmiendo')
+    # time.sleep(4)
+    # cfg.logger.error('Despertando')
     if isinstance(keys, str):
         keys = ['key', keys]
     xdotool(keys)
@@ -258,18 +276,25 @@ def open_dialog_i(cfg, name, keys, no_show=False, no_wait=False, no_main=False,
     if isinstance(name, str):
         name = [name]
     name_w_pre = [pre_gtk+f for f in name]
+    if raise_if is not None:
+        name_w_pre.extend(raise_if)
     # Add the async dialogs
-    for t in INFO_DIALOGS:
+    for t in ASYNC_DIALOGS:
         name_w_pre.append(pre_gtk_title+t)
     # Wait for our dialog or any async dialog
     # Note: wait_queue won't dismiss them because we use "with_windows=True"
     while True:
         res = wait_queue(cfg, name_w_pre, with_windows=True)
+        if raise_if is not None and res in raise_if:
+            raise InterruptedError()
         title = res[len(pre_gtk_title):]
-        if title not in INFO_DIALOGS:
+        if title not in ASYNC_DIALOGS:
             break
-        # Get rid of the info dialog
-        dismiss_pcb_info(cfg, title)
+        if title in INFO_DIALOGS:
+            # Get rid of the info dialog
+            dismiss_pcb_info(cfg, title)
+        else:
+            dismiss_pcbnew_warning(cfg, title)
         # Send the keys again
         xdotool(keys)
     name = res[len(pre_gtk):]
@@ -291,7 +316,7 @@ def check_text_replace(cfg, name):
 def paste_text_i(cfg, msg, text):
     """ Paste some text and check the echo from KiCad, then wait for sleep """
     # Paste the name
-    cfg.logger.info(msg)
+    cfg.logger.info('{} ({})'.format(msg, text))
     wait_point(cfg)
     retry = True
     while retry:
@@ -330,7 +355,7 @@ def setup_interposer_filename(cfg, fn=None):
         os.remove(BOGUS_FILENAME)
 
 
-def send_keys(cfg, msg, keys, closes=None, delay_io=False):
+def send_keys(cfg, msg, keys, closes=None, delay_io=False, no_destroy=False):
     cfg.logger.info(msg)
     wait_point(cfg)
     if isinstance(keys, str):
@@ -339,14 +364,21 @@ def send_keys(cfg, msg, keys, closes=None, delay_io=False):
         collect_io_from_queue(cfg)
     xdotool(keys)
     if closes is not None:
-        if isinstance(closes, str):
-            closes = [closes]
-        for w in closes:
-            wait_queue(cfg, 'GTK:Window Destroy:'+w)
+        if no_destroy:
+            wait_close_dialog_i(cfg)
+        else:
+            if isinstance(closes, str):
+                closes = [closes]
+            for w in closes:
+                try:
+                    wait_queue(cfg, 'GTK:Window Destroy:'+w, dialog_interrupts=True)
+                except InterruptedError:
+                    # KiCad 7.99
+                    xdotool(keys)
         wait_kicad_ready_i(cfg)
 
 
-def wait_create_i(cfg, name, fn=None):
+def wait_create_i(cfg, name, fn=None, forced_ext=None):
     """ Wait for open+close of the file.
         Also look for them in the collected_io messages.
         And if we just get close forget about the open. """
@@ -354,8 +386,16 @@ def wait_create_i(cfg, name, fn=None):
     wait_point(cfg)
     if fn is None:
         fn = cfg.output_file
-    open_msg = 'IO:open:'+fn
-    close_msg = 'IO:close:'+fn
+    fn_kicad = fn+'.'+forced_ext if forced_ext and cfg.ki8 else fn
+    # Experimental option to use the PID approach
+    # Could help for VirtioFS where the close seems to be somehow bypassed
+    use_pid = os.environ.get('KIAUTO_USE_PID_FOR_CREATE')
+    if use_pid is not None and use_pid != '0':
+        pid = cfg.pcbnew_pid if hasattr(cfg, 'pcbnew_pid') else cfg.eeschema_pid
+        return wait_for_file_created_by_process(pid, fn_kicad)
+    # Normal mechanism using the interposer
+    open_msg = 'IO:open:'+fn_kicad
+    close_msg = 'IO:close:'+fn_kicad
     if cfg.collecting_io:
         cfg.collecting_io = False
         got_open = open_msg in cfg.collected_io
@@ -374,6 +414,8 @@ def wait_create_i(cfg, name, fn=None):
     else:
         wait_queue(cfg, close_msg, starts=True)
     wait_kicad_ready_i(cfg)
+    if forced_ext and cfg.ki8 and os.path.isfile(fn_kicad):
+        os.rename(fn_kicad, fn)
 
 
 def collect_dialog_messages(cfg, title):
@@ -396,6 +438,15 @@ def exit_pcb_ees_error(cfg):
     exit(PCBNEW_ERROR if cfg.is_pcbnew else EESCHEMA_ERROR)
 
 
+def trace_dump(cfg):
+    cfg.logger.error('Trace stack:')
+    (type, value, traceback) = exc_info()
+    if traceback is None:
+        print(''.join(format_list(extract_stack()[:-2])))
+    else:
+        print_tb(traceback)
+
+
 def unknown_dialog(cfg, title, msgs=None, fatal=True):
     if msgs is None:
         msgs = collect_dialog_messages(cfg, title)
@@ -404,6 +455,7 @@ def unknown_dialog(cfg, title, msgs=None, fatal=True):
     if fatal:
         cfg.logger.error(msg_unk)
         cfg.logger.error(msg_msgs)
+        trace_dump(cfg)
         exit_pcb_ees_error(cfg)
     cfg.logger.warning(msg_unk)
     cfg.logger.warning(msg_msgs)
@@ -446,6 +498,12 @@ def dismiss_error(cfg, title):
         # KiCad 6 loading a sheet, but sub-sheets are missing
         cfg.logger.error('Error loading schematic file. Missing schematic files?')
         exit(EESCHEMA_ERROR)
+    if ("Failed to inspect the lock file '/tmp/org.kicad.kicad/instances/pcbnew-8.0' (error 2: No such file or directory)"
+       in msgs):
+        # KiCad 8.0.0RC2 glitch
+        cfg.logger.warning('KiCad 8 lock glitch')
+        dismiss_dialog(cfg, title, 'Return')
+        return
     for msg in msgs:
         if msg.startswith("Error loading schematic '"+cfg.input_file+"'."):
             # KiCad 7 schematic loading error
@@ -459,7 +517,19 @@ def dismiss_file_open_error(cfg, title):
     msgs = collect_dialog_messages(cfg, title)
     kind = 'PCB' if cfg.is_pcbnew else 'Schematic'
     fname = os.path.basename(cfg.input_file)
-    if 'Open Anyway' in msgs and kind+" '"+fname+"' is already open." in msgs:
+    # KiCad 6.x and <7.0.7: PCB 'xxxx' is already open.
+    # KiCad 7.0.7: PCB 'xxxx' is already open by 'user' at 'host'
+    # KiCad 8.0.1: PCB 'ABSOLUTE/xxxx' is already open by 'user' at 'host'
+    start = kind+" '"
+    follow = "' is already open"
+    found = False
+    cfg.logger.error(start)
+    for msg in msgs:
+        if msg.startswith(start) and follow in msg and msg.endswith("."):
+            found = True
+            fname = msg
+            break
+    if 'Open Anyway' in msgs and found:
         cfg.logger.warning('This file is already opened ({})'.format(fname))
         dismiss_dialog(cfg, title, ['Left', 'Return'])
         return
@@ -492,12 +562,28 @@ def dismiss_warning(cfg, title):
 
 
 def dismiss_pcbnew_warning(cfg, title):
-    """ Pad in invalid layer """
+    """ Pad in invalid layer
+        Missing font """
     msgs = collect_dialog_messages(cfg, title)
     # More generic cases
     for msg in msgs:
         # Warning about pad using an invalid layer
-        if msg.endswith("could not find valid layer for pad"):
+        # Missing font
+        if msg.endswith("could not find valid layer for pad") or \
+           re.search(r"Font '(.*)' not found; substituting '(.*)'", msg) or \
+           msg.startswith("Altium layer"):
+            cfg.logger.warning(msg)
+            dismiss_dialog(cfg, title, 'Return')
+            return
+    unknown_dialog(cfg, title, msgs)
+
+
+def dismiss_pcbnew_error(cfg, title):
+    """ lock glitch in 8.0.0RC2 """
+    msgs = collect_dialog_messages(cfg, title)
+    # More generic cases
+    for msg in msgs:
+        if msg.startswith("Failed to inspect the lock file"):
             cfg.logger.warning(msg)
             dismiss_dialog(cfg, title, 'Return')
             return
@@ -575,11 +661,11 @@ def exit_kicad_i(cfg):
         xdotool(['key', 'ctrl+q'])
 
 
-# def wait_close_dialog_i(cfg):
-#     """ Wait for the end of the main loop for the dialog.
-#         Then the main loop for the parent exits and enters again. """
-#     wait_queue(cfg, 'GTK:Main:Out', times=2)
-#     wait_queue(cfg, 'GTK:Main:In')
+def wait_close_dialog_i(cfg):
+    """ Wait for the end of the main loop for the dialog.
+        Then the main loop for the parent exits and enters again. """
+    wait_queue(cfg, 'GTK:Main:Out')
+    wait_queue(cfg, 'GTK:Main:In')
 
 
 def wait_and_show_progress(cfg, msg, regex_str, trigger, msg_reg, skip_match=None, with_windows=False):
@@ -614,13 +700,6 @@ def wait_and_show_progress(cfg, msg, regex_str, trigger, msg_reg, skip_match=Non
 
 
 def wait_start_by_msg(cfg):
-    cfg.logger.info('Waiting for PCB new window ...')
-    pre = 'GTK:Window Title:'
-    pre_l = len(pre)
-    cfg.logger.debug('Waiting pcbnew to start and load the PCB')
-    # Inform the elapsed time for slow loads
-    pres = [pre, 'PANGO:0:']
-    elapsed_r = re.compile(r'PANGO:(\d:\d\d:\d\d)')
     if cfg.is_pcbnew:
         kind = 'PCB'
         prg_name = 'Pcbnew'
@@ -629,13 +708,20 @@ def wait_start_by_msg(cfg):
         kind = 'Schematic'
         prg_name = 'Eeschema'
         unsaved = ' noname.sch'
+    cfg.logger.info('Waiting for {} window ...'.format(prg_name))
+    pre = 'GTK:Window Title:'
+    pre_l = len(pre)
+    cfg.logger.debug('Waiting {} to start and load the {}'.format(prg_name, kind))
+    # Inform the elapsed time for slow loads
+    pres = [pre, 'PANGO:0:']
+    elapsed_r = re.compile(r'PANGO:(\d:\d\d:\d\d)')
     loading_msg = 'Loading '+kind
     prg_msg = prg_name+' —'
     with_elapsed = False
     while True:
         # Wait for any window
         res = wait_queue(cfg, pres, starts=True, timeout=cfg.wait_start, with_windows=True)
-        cfg.logger.debug('wait_pcbew_start_by_msg got '+res)
+        cfg.logger.debug('wait_start_by_msg got '+res)
         match = elapsed_r.match(res)
         title = res[pre_l:]
         if not match and with_elapsed:
@@ -682,11 +768,17 @@ def wait_start_by_msg(cfg):
             dismiss_already_running(cfg, title)
         elif title == 'Warning':
             dismiss_warning(cfg, title)
-        elif title == 'pcbnew Warning':
+        elif title == 'pcbnew Warning' or title in WARN_DIALOGS:
             dismiss_pcbnew_warning(cfg, title)
         elif title == 'Remap Symbols':
             dismiss_remap_symbols(cfg, title)
         elif title in INFO_DIALOGS:
             dismiss_pcb_info(cfg, title)
+        elif title.startswith(KIKIT_HIDE):
+            # Buggy KiKit plugin creating a dialog at start-up (many times)
+            pass
+        elif title == 'Report':
+            # KiCad 8.0.3 bogus hidden dialog
+            pass
         else:
             unknown_dialog(cfg, title)
